@@ -4,17 +4,18 @@
 
 // https://drafts.csswg.org/css-syntax/#parsing
 
+use compact_cow_str::CompactCowStr;
+use parser::{parse_until_before, parse_until_after, parse_nested_block};
 use std::ascii::AsciiExt;
 use std::ops::Range;
-use std::borrow::Cow;
-use super::{Token, Parser, Delimiter, SourcePosition};
+use super::{Token, Parser, Delimiter, SourcePosition, ParseError, BasicParseError};
 
 
 /// Parse `!important`.
 ///
 /// Typical usage is `input.try(parse_important).is_ok()`
 /// at the end of a `DeclarationParser::parse_value` implementation.
-pub fn parse_important(input: &mut Parser) -> Result<(), ()> {
+pub fn parse_important<'i, 't>(input: &mut Parser<'i, 't>) -> Result<(), BasicParseError<'i>> {
     try!(input.expect_delim('!'));
     input.expect_ident_matching("important")
 }
@@ -43,15 +44,17 @@ pub enum AtRuleType<P, R> {
     OptionalBlock(P),
 }
 
-
 /// A trait to provide various parsing of declaration values.
 ///
 /// For example, there could be different implementations for property declarations in style rules
 /// and for descriptors in `@font-face` rules.
-pub trait DeclarationParser {
+pub trait DeclarationParser<'i> {
     /// The finished representation of a declaration.
     type Declaration;
 
+    /// The error type that is included in the ParseError value that can be returned.
+    type Error: 'i;
+
     /// Parse the value of a declaration with the given `name`.
     ///
     /// Return the finished representation for the declaration
@@ -69,10 +72,10 @@ pub trait DeclarationParser {
     /// If `!important` can be used in a given context,
     /// `input.try(parse_important).is_ok()` should be used at the end
     /// of the implementation of this method and the result should be part of the return value.
-    fn parse_value(&self, name: &str, input: &mut Parser) -> Result<Self::Declaration, ()>;
+    fn parse_value<'t>(&mut self, name: CompactCowStr<'i>, input: &mut Parser<'i, 't>)
+                       -> Result<Self::Declaration, ParseError<'i, Self::Error>>;
 }
 
-
 /// A trait to provide various parsing of at-rules.
 ///
 /// For example, there could be different implementations for top-level at-rules
@@ -82,13 +85,16 @@ pub trait DeclarationParser {
 /// Default implementations that reject all at-rules are provided,
 /// so that `impl AtRuleParser<(), ()> for ... {}` can be used
 /// for using `DeclarationListParser` to parse a declartions list with only qualified rules.
-pub trait AtRuleParser {
+pub trait AtRuleParser<'i> {
     /// The intermediate representation of an at-rule prelude.
     type Prelude;
 
     /// The finished representation of an at-rule.
     type AtRule;
 
+    /// The error type that is included in the ParseError value that can be returned.
+    type Error: 'i;
+
     /// Parse the prelude of an at-rule with the given `name`.
     ///
     /// Return the representation of the prelude and the type of at-rule,
@@ -106,11 +112,11 @@ pub trait AtRuleParser {
     /// The given `input` is a "delimited" parser
     /// that ends wherever the prelude should end.
     /// (Before the next semicolon, the next `{`, or the end of the current block.)
-    fn parse_prelude(&self, name: &str, input: &mut Parser)
-                     -> Result<AtRuleType<Self::Prelude, Self::AtRule>, ()> {
+    fn parse_prelude<'t>(&mut self, name: CompactCowStr<'i>, input: &mut Parser<'i, 't>)
+                     -> Result<AtRuleType<Self::Prelude, Self::AtRule>, ParseError<'i, Self::Error>> {
         let _ = name;
         let _ = input;
-        Err(())
+        Err(ParseError::Basic(BasicParseError::AtRuleInvalid))
     }
 
     /// Parse the content of a `{ /* ... */ }` block for the body of the at-rule.
@@ -121,25 +127,24 @@ pub trait AtRuleParser {
     ///
     /// This is only called when `parse_prelude` returned `WithBlock` or `OptionalBlock`,
     /// and a block was indeed found following the prelude.
-    fn parse_block(&self, prelude: Self::Prelude, input: &mut Parser)
-                   -> Result<Self::AtRule, ()> {
+    fn parse_block<'t>(&mut self, prelude: Self::Prelude, input: &mut Parser<'i, 't>)
+                       -> Result<Self::AtRule, ParseError<'i, Self::Error>> {
         let _ = prelude;
         let _ = input;
-        Err(())
+        Err(ParseError::Basic(BasicParseError::AtRuleInvalid))
     }
 
     /// An `OptionalBlock` prelude was followed by `;`.
     ///
     /// Convert the prelude into the finished representation of the at-rule
     /// as returned by `RuleListParser::next` or `DeclarationListParser::next`.
-    fn rule_without_block(&self, prelude: Self::Prelude) -> Self::AtRule {
+    fn rule_without_block(&mut self, prelude: Self::Prelude) -> Self::AtRule {
         let _ = prelude;
         panic!("The `AtRuleParser::rule_without_block` method must be overriden \
                 if `AtRuleParser::parse_prelude` ever returns `AtRuleType::OptionalBlock`.")
     }
 }
 
-
 /// A trait to provide various parsing of qualified rules.
 ///
 /// For example, there could be different implementations
@@ -150,13 +155,16 @@ pub trait AtRuleParser {
 /// so that `impl QualifiedRuleParser<(), ()> for ... {}` can be used
 /// for example for using `RuleListParser` to parse a rule list with only at-rules
 /// (such as inside `@font-feature-values`).
-pub trait QualifiedRuleParser {
+pub trait QualifiedRuleParser<'i> {
     /// The intermediate representation of a qualified rule prelude.
     type Prelude;
 
     /// The finished representation of a qualified rule.
     type QualifiedRule;
 
+    /// The error type that is included in the ParseError value that can be returned.
+    type Error: 'i;
+
     /// Parse the prelude of a qualified rule. For style rules, this is as Selector list.
     ///
     /// Return the representation of the prelude,
@@ -166,9 +174,10 @@ pub trait QualifiedRuleParser {
     ///
     /// The given `input` is a "delimited" parser
     /// that ends where the prelude should end (before the next `{`).
-    fn parse_prelude(&self, input: &mut Parser) -> Result<Self::Prelude, ()> {
+    fn parse_prelude<'t>(&mut self, input: &mut Parser<'i, 't>)
+                         -> Result<Self::Prelude, ParseError<'i, Self::Error>> {
         let _ = input;
-        Err(())
+        Err(ParseError::Basic(BasicParseError::QualifiedRuleInvalid))
     }
 
     /// Parse the content of a `{ /* ... */ }` block for the body of the qualified rule.
@@ -176,11 +185,11 @@ pub trait QualifiedRuleParser {
     /// Return the finished representation of the qualified rule
     /// as returned by `RuleListParser::next`,
     /// or `Err(())` to ignore the entire at-rule as invalid.
-    fn parse_block(&self, prelude: Self::Prelude, input: &mut Parser)
-                   -> Result<Self::QualifiedRule, ()> {
+    fn parse_block<'t>(&mut self, prelude: Self::Prelude, input: &mut Parser<'i, 't>)
+                       -> Result<Self::QualifiedRule, ParseError<'i, Self::Error>> {
         let _ = prelude;
         let _ = input;
-        Err(())
+        Err(ParseError::Basic(BasicParseError::QualifiedRuleInvalid))
     }
 }
 
@@ -195,8 +204,9 @@ pub struct DeclarationListParser<'i: 't, 't: 'a, 'a, P> {
 }
 
 
-impl<'i, 't, 'a, I, P> DeclarationListParser<'i, 't, 'a, P>
-where P: DeclarationParser<Declaration = I> + AtRuleParser<AtRule = I> {
+impl<'i: 't, 't: 'a, 'a, I, P, E: 'i> DeclarationListParser<'i, 't, 'a, P>
+where P: DeclarationParser<'i, Declaration = I, Error = E> +
+         AtRuleParser<'i, AtRule = I, Error = E> {
     /// Create a new `DeclarationListParser` for the given `input` and `parser`.
     ///
     /// Note that all CSS declaration lists can on principle contain at-rules.
@@ -221,11 +231,12 @@ where P: DeclarationParser<Declaration = I> + AtRuleParser<AtRule = I> {
 
 /// `DeclarationListParser` is an iterator that yields `Ok(_)` for a valid declaration or at-rule
 /// or `Err(())` for an invalid one.
-impl<'i, 't, 'a, I, P> Iterator for DeclarationListParser<'i, 't, 'a, P>
-where P: DeclarationParser<Declaration = I> + AtRuleParser<AtRule = I> {
-    type Item = Result<I, Range<SourcePosition>>;
+impl<'i: 't, 't: 'a, 'a, I, P, E: 'i> Iterator for DeclarationListParser<'i, 't, 'a, P>
+where P: DeclarationParser<'i, Declaration = I, Error = E> +
+         AtRuleParser<'i, AtRule = I, Error = E> {
+    type Item = Result<I, PreciseParseError<'i, E>>;
 
-    fn next(&mut self) -> Option<Result<I, Range<SourcePosition>>> {
+    fn next(&mut self) -> Option<Result<I, PreciseParseError<'i, E>>> {
         loop {
             let start_position = self.input.position();
             match self.input.next_including_whitespace_and_comments() {
@@ -233,20 +244,28 @@ where P: DeclarationParser<Declaration = I> + AtRuleParser<AtRule = I> {
                 Ok(Token::Ident(name)) => {
                     return Some({
                         let parser = &mut self.parser;
-                        self.input.parse_until_after(Delimiter::Semicolon, |input| {
+                        // FIXME: https://github.com/rust-lang/rust/issues/42508
+                        parse_until_after::<'i, 't, _, _, _>(self.input, Delimiter::Semicolon, |input| {
                             try!(input.expect_colon());
-                            parser.parse_value(&*name, input)
+                            parser.parse_value(name, input)
                         })
-                    }.map_err(|()| start_position..self.input.position()))
+                    }.map_err(|e| PreciseParseError {
+                        error: e,
+                        span: start_position..self.input.position()
+                    }))
                 }
                 Ok(Token::AtKeyword(name)) => {
                     return Some(parse_at_rule(start_position, name, self.input, &mut self.parser))
                 }
                 Ok(_) => {
-                    return Some(self.input.parse_until_after(Delimiter::Semicolon, |_| Err(()))
-                                .map_err(|()| start_position..self.input.position()))
+                    return Some(self.input.parse_until_after(Delimiter::Semicolon,
+                                                             |_| Err(ParseError::Basic(BasicParseError::ExpectedToken(Token::Semicolon))))
+                                .map_err(|e| PreciseParseError {
+                                    error: e,
+                                    span: start_position..self.input.position()
+                                }))
                 }
-                Err(()) => return None,
+                Err(_) => return None,
             }
         }
     }
@@ -266,8 +285,9 @@ pub struct RuleListParser<'i: 't, 't: 'a, 'a, P> {
 }
 
 
-impl<'i: 't, 't: 'a, 'a, R, P> RuleListParser<'i, 't, 'a, P>
-where P: QualifiedRuleParser<QualifiedRule = R> + AtRuleParser<AtRule = R> {
+impl<'i: 't, 't: 'a, 'a, R, P, E: 'i> RuleListParser<'i, 't, 'a, P>
+where P: QualifiedRuleParser<'i, QualifiedRule = R, Error = E> +
+         AtRuleParser<'i, AtRule = R, Error = E> {
     /// Create a new `RuleListParser` for the given `input` at the top-level of a stylesheet
     /// and the given `parser`.
     ///
@@ -306,11 +326,12 @@ where P: QualifiedRuleParser<QualifiedRule = R> + AtRuleParser<AtRule = R> {
 
 
 /// `RuleListParser` is an iterator that yields `Ok(_)` for a rule or `Err(())` for an invalid one.
-impl<'i, 't, 'a, R, P> Iterator for RuleListParser<'i, 't, 'a, P>
-where P: QualifiedRuleParser<QualifiedRule = R> + AtRuleParser<AtRule = R> {
-    type Item = Result<R, Range<SourcePosition>>;
+impl<'i: 't, 't: 'a, 'a, R, P, E: 'i> Iterator for RuleListParser<'i, 't, 'a, P>
+where P: QualifiedRuleParser<'i, QualifiedRule = R, Error = E> +
+         AtRuleParser<'i, AtRule = R, Error = E> {
+    type Item = Result<R, PreciseParseError<'i, E>>;
 
-    fn next(&mut self) -> Option<Result<R, Range<SourcePosition>>> {
+    fn next(&mut self) -> Option<Result<R, PreciseParseError<'i, E>>> {
         loop {
             let start_position = self.input.position();
             match self.input.next_including_whitespace_and_comments() {
@@ -321,7 +342,7 @@ where P: QualifiedRuleParser<QualifiedRule = R> + AtRuleParser<AtRule = R> {
                     self.any_rule_so_far = true;
                     if first_stylesheet_rule && name.eq_ignore_ascii_case("charset") {
                         let delimiters = Delimiter::Semicolon | Delimiter::CurlyBracketBlock;
-                        let _ = self.input.parse_until_after(delimiters, |_input| Ok(()));
+                        let _: Result<(), ParseError<()>> = self.input.parse_until_after(delimiters, |_| Ok(()));
                     } else {
                         return Some(parse_at_rule(start_position, name, self.input, &mut self.parser))
                     }
@@ -330,9 +351,12 @@ where P: QualifiedRuleParser<QualifiedRule = R> + AtRuleParser<AtRule = R> {
                     self.any_rule_so_far = true;
                     self.input.reset(start_position);
                     return Some(parse_qualified_rule(self.input, &mut self.parser)
-                                .map_err(|()| start_position..self.input.position()))
+                                .map_err(|e| PreciseParseError {
+                                    error: e,
+                                    span: start_position..self.input.position()
+                                }))
                 }
-                Err(()) => return None,
+                Err(_) => return None,
             }
         }
     }
@@ -340,99 +364,134 @@ where P: QualifiedRuleParser<QualifiedRule = R> + AtRuleParser<AtRule = R> {
 
 
 /// Parse a single declaration, such as an `( /* ... */ )` parenthesis in an `@supports` prelude.
-pub fn parse_one_declaration<P>(input: &mut Parser, parser: &mut P)
-                                -> Result<<P as DeclarationParser>::Declaration,
-                                          Range<SourcePosition>>
-                                where P: DeclarationParser {
+pub fn parse_one_declaration<'i, 't, P, E>(input: &mut Parser<'i, 't>, parser: &mut P)
+                                           -> Result<<P as DeclarationParser<'i>>::Declaration,
+                                                     PreciseParseError<'i, E>>
+                                           where P: DeclarationParser<'i, Error = E> {
     let start_position = input.position();
     input.parse_entirely(|input| {
         let name = try!(input.expect_ident());
         try!(input.expect_colon());
-        parser.parse_value(&*name, input)
-    }).map_err(|()| start_position..input.position())
+        parser.parse_value(name, input)
+    }).map_err(|e| PreciseParseError {
+        error: e,
+        span: start_position..input.position()
+    })
 }
 
 
 /// Parse a single rule, such as for CSSOM’s `CSSStyleSheet.insertRule`.
-pub fn parse_one_rule<R, P>(input: &mut Parser, parser: &mut P) -> Result<R, ()>
-where P: QualifiedRuleParser<QualifiedRule = R> + AtRuleParser<AtRule = R> {
+pub fn parse_one_rule<'i, 't, R, P, E>(input: &mut Parser<'i, 't>, parser: &mut P)
+                                       -> Result<R, ParseError<'i, E>>
+where P: QualifiedRuleParser<'i, QualifiedRule = R, Error = E> +
+         AtRuleParser<'i, AtRule = R, Error = E> {
     input.parse_entirely(|input| {
         loop {
             let start_position = input.position();
             match try!(input.next_including_whitespace_and_comments()) {
                 Token::WhiteSpace(_) | Token::Comment(_) => {}
                 Token::AtKeyword(name) => {
-                    return parse_at_rule(start_position, name, input, parser).map_err(|_| ())
+                    return parse_at_rule(start_position, name, input, parser).map_err(|e| e.error)
                 }
                 _ => {
                     input.reset(start_position);
-                    return parse_qualified_rule(input, parser).map_err(|_| ())
+                    return parse_qualified_rule(input, parser)
                 }
             }
         }
     })
 }
 
+pub struct PreciseParseError<'i, E: 'i> {
+    pub error: ParseError<'i, E>,
+    pub span: Range<SourcePosition>,
+}
 
-fn parse_at_rule<P>(start_position: SourcePosition, name: Cow<str>,
-                    input: &mut Parser, parser: &mut P)
-                    -> Result<<P as AtRuleParser>::AtRule, Range<SourcePosition>>
-                    where P: AtRuleParser {
+fn parse_at_rule<'i: 't, 't, P, E>(start_position: SourcePosition, name: CompactCowStr<'i>,
+                                   input: &mut Parser<'i, 't>, parser: &mut P)
+                                   -> Result<<P as AtRuleParser<'i>>::AtRule, PreciseParseError<'i, E>>
+                                   where P: AtRuleParser<'i, Error = E> {
     let delimiters = Delimiter::Semicolon | Delimiter::CurlyBracketBlock;
-    let result = input.parse_until_before(delimiters, |input| {
-        parser.parse_prelude(&*name, input)
+    // FIXME: https://github.com/rust-lang/rust/issues/42508
+    let result = parse_until_before::<'i, 't, _, _, _>(input, delimiters, |input| {
+        parser.parse_prelude(name, input)
     });
     match result {
         Ok(AtRuleType::WithoutBlock(rule)) => {
             match input.next() {
-                Ok(Token::Semicolon) | Err(()) => Ok(rule),
-                Ok(Token::CurlyBracketBlock) => Err(start_position..input.position()),
+                Ok(Token::Semicolon) | Err(_) => Ok(rule),
+                Ok(Token::CurlyBracketBlock) => Err(PreciseParseError {
+                    error: ParseError::Basic(BasicParseError::UnexpectedToken(Token::CurlyBracketBlock)),
+                    span: start_position..input.position(),
+                }),
                 Ok(_) => unreachable!()
             }
         }
         Ok(AtRuleType::WithBlock(prelude)) => {
             match input.next() {
                 Ok(Token::CurlyBracketBlock) => {
-                    input.parse_nested_block(move |input| parser.parse_block(prelude, input))
-                    .map_err(|()| start_position..input.position())
+                    // FIXME: https://github.com/rust-lang/rust/issues/42508
+                    parse_nested_block::<'i, 't, _, _, _>(input, move |input| parser.parse_block(prelude, input))
+                        .map_err(|e| PreciseParseError {
+                            error: e,
+                            span: start_position..input.position(),
+                        })
                 }
-                Ok(Token::Semicolon) | Err(()) => Err(start_position..input.position()),
+                Ok(Token::Semicolon) => Err(PreciseParseError {
+                    error: ParseError::Basic(BasicParseError::UnexpectedToken(Token::Semicolon)),
+                    span: start_position..input.position()
+                }),
+                Err(e) => Err(PreciseParseError {
+                    error: ParseError::Basic(e),
+                    span: start_position..input.position(),
+                }),
                 Ok(_) => unreachable!()
             }
         }
         Ok(AtRuleType::OptionalBlock(prelude)) => {
             match input.next() {
-                Ok(Token::Semicolon) | Err(()) => Ok(parser.rule_without_block(prelude)),
+                Ok(Token::Semicolon) | Err(_) => Ok(parser.rule_without_block(prelude)),
                 Ok(Token::CurlyBracketBlock) => {
-                    input.parse_nested_block(move |input| parser.parse_block(prelude, input))
-                    .map_err(|()| start_position..input.position())
+                    // FIXME: https://github.com/rust-lang/rust/issues/42508
+                    parse_nested_block::<'i, 't, _, _, _>(input, move |input| parser.parse_block(prelude, input))
+                        .map_err(|e| PreciseParseError {
+                            error: e,
+                            span: start_position..input.position(),
+                        })
                 }
                 _ => unreachable!()
             }
         }
-        Err(()) => {
+        Err(_) => {
             let end_position = input.position();
-            match input.next() {
-                Ok(Token::CurlyBracketBlock) | Ok(Token::Semicolon) | Err(()) => {}
+            let error = match input.next() {
+                Ok(Token::CurlyBracketBlock) => BasicParseError::UnexpectedToken(Token::CurlyBracketBlock),
+                Ok(Token::Semicolon) => BasicParseError::UnexpectedToken(Token::Semicolon),
+                Err(e) => e,
                 _ => unreachable!()
-            }
-            Err(start_position..end_position)
+            };
+            Err(PreciseParseError {
+                error: ParseError::Basic(error),
+                span: start_position..end_position,
+            })
         }
     }
 }
 
 
-fn parse_qualified_rule<P>(input: &mut Parser, parser: &mut P)
-                           -> Result<<P as QualifiedRuleParser>::QualifiedRule, ()>
-                           where P: QualifiedRuleParser {
-    let prelude = input.parse_until_before(Delimiter::CurlyBracketBlock, |input| {
+fn parse_qualified_rule<'i, 't, P, E>(input: &mut Parser<'i, 't>, parser: &mut P)
+                                      -> Result<<P as QualifiedRuleParser<'i>>::QualifiedRule, ParseError<'i, E>>
+                                      where P: QualifiedRuleParser<'i, Error = E> {
+    // FIXME: https://github.com/rust-lang/rust/issues/42508
+    let prelude = parse_until_before::<'i, 't, _, _, _>(input, Delimiter::CurlyBracketBlock, |input| {
         parser.parse_prelude(input)
     });
     match try!(input.next()) {
         Token::CurlyBracketBlock => {
             // Do this here so that we consume the `{` even if the prelude is `Err`.
             let prelude = try!(prelude);
-            input.parse_nested_block(move |input| parser.parse_block(prelude, input))
+            // FIXME: https://github.com/rust-lang/rust/issues/42508
+            parse_nested_block::<'i, 't, _, _, _>(input, move |input| parser.parse_block(prelude, input))
         }
         _ => unreachable!()
     }
