@@ -6,23 +6,17 @@
 extern crate test;
 
 use encoding_rs;
-use std::borrow::Cow::{self, Borrowed};
-use std::fs::File;
-use std::io::{self, Write};
-use std::path::Path;
-use std::process::Command;
 use rustc_serialize::json::{self, Json, ToJson};
-use tempdir::TempDir;
 
 #[cfg(feature = "bench")]
 use self::test::Bencher;
 
-use super::{Parser, Delimiter, Token, NumericValue, PercentageValue, SourceLocation,
-            DeclarationListParser, DeclarationParser, RuleListParser,
-            AtRuleType, AtRuleParser, QualifiedRuleParser,
+use super::{Parser, Delimiter, Token, SourceLocation, ParseError,
+            DeclarationListParser, DeclarationParser, RuleListParser, BasicParseError,
+            AtRuleType, AtRuleParser, QualifiedRuleParser, ParserInput,
             parse_one_declaration, parse_one_rule, parse_important,
             stylesheet_encoding, EncodingSupport,
-            TokenSerializationType,
+            TokenSerializationType, CompactCowStr,
             Color, RGBA, parse_nth, UnicodeRange, ToCss};
 
 macro_rules! JArray {
@@ -30,29 +24,6 @@ macro_rules! JArray {
     ($($e: expr),*) => { Json::Array(vec!( $( $e.to_json() ),* )) }
 }
 
-
-fn write_whole_file(path: &Path, data: &str) -> io::Result<()> {
-    (try!(File::create(path))).write_all(data.as_bytes())
-}
-
-
-fn print_json_diff(results: &Json, expected: &Json) -> io::Result<()> {
-    let temp = try!(TempDir::new("rust-cssparser-tests"));
-    let results = results.pretty().to_string() + "\n";
-    let expected = expected.pretty().to_string() + "\n";
-    let result_path = temp.path().join("results.json");
-    let expected_path = temp.path().join("expected.json");
-    try!(write_whole_file(&result_path, &results));
-    try!(write_whole_file(&expected_path, &expected));
-    try!(Command::new("colordiff")
-        .arg("-u1000")
-        .arg(&result_path)
-        .arg(&expected_path)
-        .status());
-    Ok(())
-}
-
-
 fn almost_equals(a: &Json, b: &Json) -> bool {
     match (a, b) {
         (&Json::I64(a), _) => almost_equals(&Json::F64(a as f64), b),
@@ -90,11 +61,15 @@ fn normalize(json: &mut Json) {
     }
 }
 
-fn assert_json_eq(results: json::Json, mut expected: json::Json, message: String) {
+fn assert_json_eq(results: json::Json, mut expected: json::Json, message: &str) {
     normalize(&mut expected);
     if !almost_equals(&results, &expected) {
-        print_json_diff(&results, &expected).unwrap();
-        panic!(message)
+        println!("{}", ::difference::Changeset::new(
+            &results.pretty().to_string(),
+            &expected.pretty().to_string(),
+            "\n",
+        ));
+        panic!("{}", message)
     }
 }
 
@@ -121,8 +96,9 @@ fn run_json_tests<F: Fn(&mut Parser) -> Json>(json_data: &str, parse: F) {
     run_raw_json_tests(json_data, |input, expected| {
         match input {
             Json::String(input) => {
-                let result = parse(&mut Parser::new(&input));
-                assert_json_eq(result, expected, input);
+                let mut parse_input = ParserInput::new(&input);
+                let result = parse(&mut Parser::new(&mut parse_input));
+                assert_json_eq(result, expected, &input);
             },
             _ => panic!("Unexpected JSON")
         }
@@ -141,9 +117,10 @@ fn component_value_list() {
 #[test]
 fn one_component_value() {
     run_json_tests(include_str!("css-parsing-tests/one_component_value.json"), |input| {
-        input.parse_entirely(|input| {
+        let result: Result<Json, ParseError<()>> = input.parse_entirely(|input| {
             Ok(one_component_value_to_json(try!(input.next()), input))
-        }).unwrap_or(JArray!["error", "invalid"])
+        });
+        result.unwrap_or(JArray!["error", "invalid"])
     });
 }
 
@@ -237,13 +214,14 @@ fn stylesheet_from_bytes() {
             let encoding = stylesheet_encoding::<EncodingRs>(
                 &css, protocol_encoding_label, environment_encoding);
             let (css_unicode, used_encoding, _) = encoding.decode(&css);
-            let input = &mut Parser::new(&css_unicode);
+            let mut input = ParserInput::new(&css_unicode);
+            let input = &mut Parser::new(&mut input);
             let rules = RuleListParser::new_for_stylesheet(input, JsonParser)
                         .map(|result| result.unwrap_or(JArray!["error", "invalid"]))
                         .collect::<Vec<_>>();
             JArray![rules, used_encoding.name().to_lowercase()]
         };
-        assert_json_eq(result, expected, Json::Object(map).to_string());
+        assert_json_eq(result, expected, &Json::Object(map).to_string());
     });
 
     fn get_string<'a>(map: &'a json::Object, key: &str) -> Option<&'a str> {
@@ -259,25 +237,38 @@ fn stylesheet_from_bytes() {
 
 #[test]
 fn expect_no_error_token() {
-    assert!(Parser::new("foo 4px ( / { !bar }").expect_no_error_token().is_ok());
-    assert!(Parser::new(")").expect_no_error_token().is_err());
-    assert!(Parser::new("}").expect_no_error_token().is_err());
-    assert!(Parser::new("(a){]").expect_no_error_token().is_err());
-    assert!(Parser::new("'\n'").expect_no_error_token().is_err());
-    assert!(Parser::new("url('\n'").expect_no_error_token().is_err());
-    assert!(Parser::new("url(a b)").expect_no_error_token().is_err());
-    assert!(Parser::new("url(\u{7F})").expect_no_error_token().is_err());
+    let mut input = ParserInput::new("foo 4px ( / { !bar }");
+    assert!(Parser::new(&mut input).expect_no_error_token().is_ok());
+    let mut input = ParserInput::new(")");
+    assert!(Parser::new(&mut input).expect_no_error_token().is_err());
+    let mut input = ParserInput::new("}");
+    assert!(Parser::new(&mut input).expect_no_error_token().is_err());
+    let mut input = ParserInput::new("(a){]");
+    assert!(Parser::new(&mut input).expect_no_error_token().is_err());
+    let mut input = ParserInput::new("'\n'");
+    assert!(Parser::new(&mut input).expect_no_error_token().is_err());
+    let mut input = ParserInput::new("url('\n'");
+    assert!(Parser::new(&mut input).expect_no_error_token().is_err());
+    let mut input = ParserInput::new("url(a b)");
+    assert!(Parser::new(&mut input).expect_no_error_token().is_err());
+    let mut input = ParserInput::new("url(\u{7F}))");
+    assert!(Parser::new(&mut input).expect_no_error_token().is_err());
 }
 
 
 /// https://github.com/servo/rust-cssparser/issues/71
 #[test]
 fn outer_block_end_consumed() {
-    let mut input = Parser::new("(calc(true))");
+    let mut input = ParserInput::new("(calc(true))");
+    let mut input = Parser::new(&mut input);
     assert!(input.expect_parenthesis_block().is_ok());
-    assert!(input.parse_nested_block(|input| input.expect_function_matching("calc")).is_ok());
+    assert!(input.parse_nested_block(|input| {
+        let result: Result<_, ParseError<()>> = input.expect_function_matching("calc")
+            .map_err(|e| ParseError::Basic(e));
+        result
+    }).is_ok());
     println!("{:?}", input.position());
-    assert_eq!(input.next(), Err(()));
+    assert!(input.next().is_err());
 }
 
 #[test]
@@ -296,30 +287,41 @@ fn unquoted_url_escaping() {
             !\\\"#$%&\\'\\(\\)*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\\\]\
             ^_`abcdefghijklmnopqrstuvwxyz{|}~\\7F é\
         )\
-    ");
-    assert_eq!(Parser::new(&serialized).next(), Ok(token))
+        ");
+    let mut input = ParserInput::new(&serialized);
+    assert_eq!(Parser::new(&mut input).next(), Ok(token));
 }
 
 #[test]
 fn test_expect_url() {
-    fn parse(s: &str) -> Result<Cow<str>, ()> {
+    fn parse<'a>(s: &mut ParserInput<'a>) -> Result<CompactCowStr<'a>, BasicParseError<'a>> {
         Parser::new(s).expect_url()
     }
-    assert_eq!(parse("url()").unwrap(), "");
-    assert_eq!(parse("url( ").unwrap(), "");
-    assert_eq!(parse("url( abc").unwrap(), "abc");
-    assert_eq!(parse("url( abc \t)").unwrap(), "abc");
-    assert_eq!(parse("url( 'abc' \t)").unwrap(), "abc");
-    assert_eq!(parse("url(abc more stuff)"), Err(()));
+    let mut input = ParserInput::new("url()");
+    assert_eq!(parse(&mut input).unwrap(), "");
+    let mut input = ParserInput::new("url( ");
+    assert_eq!(parse(&mut input).unwrap(), "");
+    let mut input = ParserInput::new("url( abc");
+    assert_eq!(parse(&mut input).unwrap(), "abc");
+    let mut input = ParserInput::new("url( abc \t)");
+    assert_eq!(parse(&mut input).unwrap(), "abc");
+    let mut input = ParserInput::new("url( 'abc' \t)");
+    assert_eq!(parse(&mut input).unwrap(), "abc");
+    let mut input = ParserInput::new("url(abc more stuff)");
+    assert!(parse(&mut input).is_err());
     // The grammar at https://drafts.csswg.org/css-values/#urls plans for `<url-modifier>*`
     // at the position of "more stuff", but no such modifier is defined yet.
-    assert_eq!(parse("url('abc' more stuff)"), Err(()));
+    let mut input = ParserInput::new("url('abc' more stuff)");
+    assert!(parse(&mut input).is_err());
 }
 
 
 fn run_color_tests<F: Fn(Result<Color, ()>) -> Json>(json_data: &str, to_json: F) {
     run_json_tests(json_data, |input| {
-        to_json(input.parse_entirely(Color::parse))
+        let result: Result<_, ParseError<()>> = input.parse_entirely(|i| {
+            Color::parse(i).map_err(|e| ParseError::Basic(e))
+        });
+        to_json(result.map_err(|_| ()))
     });
 }
 
@@ -346,14 +348,17 @@ fn color3_keywords() {
 #[test]
 fn nth() {
     run_json_tests(include_str!("css-parsing-tests/An+B.json"), |input| {
-        input.parse_entirely(parse_nth).ok().to_json()
+        input.parse_entirely(|i| {
+            let result: Result<_, ParseError<()>> = parse_nth(i).map_err(|e| ParseError::Basic(e));
+            result
+        }).ok().to_json()
     });
 }
 
 #[test]
 fn unicode_range() {
     run_json_tests(include_str!("css-parsing-tests/urange.json"), |input| {
-        input.parse_comma_separated(|input| {
+        let result: Result<_, ParseError<()>> = input.parse_comma_separated(|input| {
             let result = UnicodeRange::parse(input).ok().map(|r| (r.start, r.end));
             if input.is_exhausted() {
                 Ok(result)
@@ -361,7 +366,8 @@ fn unicode_range() {
                 while let Ok(_) = input.next() {}
                 Ok(None)
             }
-        }).unwrap().to_json()
+        });
+        result.unwrap().to_json()
     });
 }
 
@@ -400,17 +406,19 @@ fn serializer(preserve_comments: bool) {
                     _ => None
                 };
                 if let Some(closing_token) = closing_token {
-                    input.parse_nested_block(|input| {
+                    let result: Result<_, ParseError<()>> = input.parse_nested_block(|input| {
                         write_to(previous_token, input, string, preserve_comments);
                         Ok(())
-                    }).unwrap();
+                    });
+                    result.unwrap();
                     closing_token.to_css(string).unwrap();
                 }
             }
         }
         let mut serialized = String::new();
         write_to(TokenSerializationType::nothing(), input, &mut serialized, preserve_comments);
-        let parser = &mut Parser::new(&serialized);
+        let mut input = ParserInput::new(&serialized);
+        let parser = &mut Parser::new(&mut input);
         Json::Array(component_values_to_json(parser))
     });
 }
@@ -441,17 +449,18 @@ fn serialize_rgba_two_digit_float_if_roundtrips() {
 
 #[test]
 fn line_numbers() {
-    let mut input = Parser::new("foo bar\nbaz\r\n\n\"a\\\r\nb\"");
+    let mut input = ParserInput::new("foo bar\nbaz\r\n\n\"a\\\r\nb\"");
+    let mut input = Parser::new(&mut input);
     assert_eq!(input.current_source_location(), SourceLocation { line: 1, column: 1 });
-    assert_eq!(input.next_including_whitespace(), Ok(Token::Ident(Borrowed("foo"))));
+    assert_eq!(input.next_including_whitespace(), Ok(Token::Ident("foo".into())));
     assert_eq!(input.current_source_location(), SourceLocation { line: 1, column: 4 });
     assert_eq!(input.next_including_whitespace(), Ok(Token::WhiteSpace(" ")));
     assert_eq!(input.current_source_location(), SourceLocation { line: 1, column: 5 });
-    assert_eq!(input.next_including_whitespace(), Ok(Token::Ident(Borrowed("bar"))));
+    assert_eq!(input.next_including_whitespace(), Ok(Token::Ident("bar".into())));
     assert_eq!(input.current_source_location(), SourceLocation { line: 1, column: 8 });
     assert_eq!(input.next_including_whitespace(), Ok(Token::WhiteSpace("\n")));
     assert_eq!(input.current_source_location(), SourceLocation { line: 2, column: 1 });
-    assert_eq!(input.next_including_whitespace(), Ok(Token::Ident(Borrowed("baz"))));
+    assert_eq!(input.next_including_whitespace(), Ok(Token::Ident("baz".into())));
     assert_eq!(input.current_source_location(), SourceLocation { line: 2, column: 4 });
     let position = input.position();
 
@@ -460,9 +469,9 @@ fn line_numbers() {
 
     assert_eq!(input.source_location(position), SourceLocation { line: 2, column: 4 });
 
-    assert_eq!(input.next_including_whitespace(), Ok(Token::QuotedString(Borrowed("ab"))));
+    assert_eq!(input.next_including_whitespace(), Ok(Token::QuotedString("ab".into())));
     assert_eq!(input.current_source_location(), SourceLocation { line: 5, column: 3 });
-    assert_eq!(input.next_including_whitespace(), Err(()));
+    assert!(input.next_including_whitespace().is_err());
 }
 
 #[test]
@@ -494,7 +503,8 @@ fn overflow() {
          -3.402824e+38
 
     ".replace("{309 zeros}", &repeat('0').take(309).collect::<String>());
-    let mut input = Parser::new(&css);
+    let mut input = ParserInput::new(&css);
+    let mut input = Parser::new(&mut input);
 
     assert_eq!(input.expect_integer(), Ok(2147483646));
     assert_eq!(input.expect_integer(), Ok(2147483647));
@@ -523,11 +533,15 @@ fn overflow() {
 
 #[test]
 fn line_delimited() {
-    let mut input = Parser::new(" { foo ; bar } baz;,");
+    let mut input = ParserInput::new(" { foo ; bar } baz;,");
+    let mut input = Parser::new(&mut input);
     assert_eq!(input.next(), Ok(Token::CurlyBracketBlock));
-    assert_eq!(input.parse_until_after(Delimiter::Semicolon, |_| Ok(42)), Err(()));
+    assert!({
+        let result: Result<_, ParseError<()>> = input.parse_until_after(Delimiter::Semicolon, |_| Ok(42));
+        result
+    }.is_err());
     assert_eq!(input.next(), Ok(Token::Comma));
-    assert_eq!(input.next(), Err(()));
+    assert!(input.next().is_err());
 }
 
 #[test]
@@ -620,7 +634,8 @@ const BACKGROUND_IMAGE: &'static str = include_str!("big-data-url.css");
 #[bench]
 fn unquoted_url(b: &mut Bencher) {
     b.iter(|| {
-        let mut input = Parser::new(BACKGROUND_IMAGE);
+        let mut input = ParserInput::new(BACKGROUND_IMAGE);
+        let mut input = Parser::new(&mut input);
         input.look_for_var_functions();
 
         let result = input.try(|input| input.expect_url());
@@ -632,6 +647,19 @@ fn unquoted_url(b: &mut Bencher) {
     })
 }
 
+
+#[cfg(feature = "bench")]
+#[bench]
+fn numeric(b: &mut Bencher) {
+    b.iter(|| {
+        for _ in 0..1000000 {
+            let mut input = ParserInput::new("10px");
+            let mut input = Parser::new(&mut input);
+            let _ = test::black_box(input.next());
+        }
+    })
+}
+
 struct JsonParser;
 
 #[test]
@@ -641,14 +669,17 @@ fn no_stack_overflow_multiple_nested_blocks() {
         let dup = input.clone();
         input.push_str(&dup);
     }
-    let mut input = Parser::new(&input);
+    let mut input = ParserInput::new(&input);
+    let mut input = Parser::new(&mut input);
     while let Ok(..) = input.next() { }
 }
 
-impl DeclarationParser for JsonParser {
+impl<'i> DeclarationParser<'i> for JsonParser {
     type Declaration = Json;
+    type Error = ();
 
-    fn parse_value(&mut self, name: &str, input: &mut Parser) -> Result<Json, ()> {
+    fn parse_value<'t>(&mut self, name: CompactCowStr<'i>, input: &mut Parser<'i, 't>)
+                       -> Result<Json, ParseError<'i, ()>> {
         let mut value = vec![];
         let mut important = false;
         loop {
@@ -683,12 +714,13 @@ impl DeclarationParser for JsonParser {
     }
 }
 
-impl AtRuleParser for JsonParser {
+impl<'i> AtRuleParser<'i> for JsonParser {
     type Prelude = Vec<Json>;
     type AtRule = Json;
+    type Error = ();
 
-    fn parse_prelude(&mut self, name: &str, input: &mut Parser)
-                     -> Result<AtRuleType<Vec<Json>, Json>, ()> {
+    fn parse_prelude<'t>(&mut self, name: CompactCowStr<'i>, input: &mut Parser<'i, 't>)
+                         -> Result<AtRuleType<Vec<Json>, Json>, ParseError<'i, ()>> {
         Ok(AtRuleType::OptionalBlock(vec![
             "at-rule".to_json(),
             name.to_json(),
@@ -696,7 +728,8 @@ impl AtRuleParser for JsonParser {
         ]))
     }
 
-    fn parse_block(&mut self, mut prelude: Vec<Json>, input: &mut Parser) -> Result<Json, ()> {
+    fn parse_block<'t>(&mut self, mut prelude: Vec<Json>, input: &mut Parser<'i, 't>)
+                       -> Result<Json, ParseError<'i, ()>> {
         prelude.push(Json::Array(component_values_to_json(input)));
         Ok(Json::Array(prelude))
     }
@@ -707,15 +740,17 @@ impl AtRuleParser for JsonParser {
     }
 }
 
-impl QualifiedRuleParser for JsonParser {
+impl<'i> QualifiedRuleParser<'i> for JsonParser {
     type Prelude = Vec<Json>;
     type QualifiedRule = Json;
+    type Error = ();
 
-    fn parse_prelude(&mut self, input: &mut Parser) -> Result<Vec<Json>, ()> {
+    fn parse_prelude<'t>(&mut self, input: &mut Parser<'i, 't>) -> Result<Vec<Json>, ParseError<'i, ()>> {
         Ok(component_values_to_json(input))
     }
 
-    fn parse_block(&mut self, prelude: Vec<Json>, input: &mut Parser) -> Result<Json, ()> {
+    fn parse_block<'t>(&mut self, prelude: Vec<Json>, input: &mut Parser<'i, 't>)
+                       -> Result<Json, ParseError<'i, ()>> {
         Ok(JArray![
             "qualified rule",
             prelude,
@@ -733,16 +768,23 @@ fn component_values_to_json(input: &mut Parser) -> Vec<Json> {
 }
 
 fn one_component_value_to_json(token: Token, input: &mut Parser) -> Json {
-    fn numeric(value: NumericValue) -> Vec<json::Json> {
+    fn numeric(value: f32, int_value: Option<i32>, has_sign: bool) -> Vec<json::Json> {
         vec![
-            Token::Number(value).to_css_string().to_json(),
-            match value.int_value { Some(i) => i.to_json(), None => value.value.to_json() },
-            match value.int_value { Some(_) => "integer", None => "number" }.to_json()
+            Token::Number {
+                value: value,
+                int_value: int_value,
+                has_sign: has_sign,
+            }.to_css_string().to_json(),
+            match int_value { Some(i) => i.to_json(), None => value.to_json() },
+            match int_value { Some(_) => "integer", None => "number" }.to_json()
         ]
     }
 
     fn nested(input: &mut Parser) -> Vec<Json> {
-        input.parse_nested_block(|input| Ok(component_values_to_json(input))).unwrap()
+        let result: Result<_, ParseError<()>> = input.parse_nested_block(|input| {
+            Ok(component_values_to_json(input))
+        });
+        result.unwrap()
     }
 
     match token {
@@ -755,23 +797,19 @@ fn one_component_value_to_json(token: Token, input: &mut Parser) -> Json {
         Token::Delim('\\') => "\\".to_json(),
         Token::Delim(value) => value.to_string().to_json(),
 
-        Token::Number(value) => Json::Array({
+        Token::Number { value, int_value, has_sign } => Json::Array({
             let mut v = vec!["number".to_json()];
-            v.extend(numeric(value));
+            v.extend(numeric(value, int_value, has_sign));
             v
         }),
-        Token::Percentage(PercentageValue { unit_value, int_value, has_sign }) => Json::Array({
+        Token::Percentage { unit_value, int_value, has_sign } => Json::Array({
             let mut v = vec!["percentage".to_json()];
-            v.extend(numeric(NumericValue {
-                value: unit_value * 100.,
-                int_value: int_value,
-                has_sign: has_sign,
-            }));
+            v.extend(numeric(unit_value * 100., int_value, has_sign));
             v
         }),
-        Token::Dimension(value, unit) => Json::Array({
+        Token::Dimension { value, int_value, has_sign, unit } => Json::Array({
             let mut v = vec!["dimension".to_json()];
-            v.extend(numeric(value));
+            v.extend(numeric(value, int_value, has_sign));
             v.push(unit.to_json());
             v
         }),
